#!/usr/bin/python
"""
Ringtale: A simple game framework
This is a way to organize the events of a game: a loop for storytelling.

As I tried to make more complex games, I ran into problems involving how
to get information from one part of a game to another, such as the game's
title screen, walking-around screen, and quick-travel screen. Fearing
infinite recursion from one game function calling another and vice versa,
I tried to link these "screens" by having a central game object put them
on a stack, controlled by a loop, with a message stack that they could use to
exchange information on various game or UI events. That system quickly got
confusing and buggy.

Here is what will hopefully be a more workable system. There's a central
game object with a main loop calling functions for various game states/screens,
but now the main loop actually runs constantly, farming out logic and drawing
to a pair of functions based on the current state name but staying in charge
of basic tasks like updating the screen and maintaining the framerate.

Messaging between parts of the game can best be handled, in this system, by
modifying a "game world" class that is part of a subclass of Framework.
That is, logic functions should call functions of that world object to make
changes to it such as adding characters.
"""

__author__ = "Kris Schnee"
__date__ = "2008.6.3"
__license__ = "Public Domain"

import time ## For FPS counter, not really necessary.
import pygame.time ## For framerate regulation.


class Framework:
    """The main class of the game.

    Attempts to follow the MVC Architecture:
    -Model: class World (to be created by you)
    -View: *Draw functions
    -Controller (and game logic): *Logic functions

    Extend this class with state functions (*Draw and *Logic)
    and a reference to/inclusion of a model/world.
    """
    def __init__(self,starting_state=None):
        self.logic_function = None
        self.draw_function = None
        self.setup_function = None
        self.states = []
        self.interface_stack = []
        self.notes_stack = []

        """A dictionary for storing interface elements."""
        self.interface = []
        self.interface_index = {}
        """Used for interface events."""
        self.messages = []

        """A way of storing information about the game.
        This is an alternative to storing every bit of info in a game
        world object, such as the status of a multi-part command being entered.
        It's never automatically reset and there's no error checking other
        than "get" to make sure an accessed piece of data has been set.
        """
        self.notes = {}

        if starting_state:
            self.PushState(starting_state)
            if self.setup_function:
                self.setup_function()

        ## Framerate regulation.
        self.clock = pygame.time.Clock()
        self.framerate = 20

    def Notify(self,event):
        self.messages.append(event)

    """ State Control """

    def GetStateFunctions(self,state_name):
        try:
            name = state_name + "Logic"
            logic_function = getattr(self,name)
            name = state_name + "Draw"
            draw_function = getattr(self,name)
        except:
            raise "Error: Couldn't find matching logic/draw functions for state \""+state_name+"\"."
        name = state_name + "Setup" ## Optional
        if hasattr(self,name):
            setup_function = getattr(self,name)
        else:
            setup_function = None
        return logic_function, draw_function, setup_function

    def PushState(self,state_name,passed_notes=[]):
        """Start a new game state, but revert when done."""
##        print "Pushing state '"+state_name+"'"

        ## Store interface elements & notes, passing some notes "up."
        self.interface_stack.append(self.interface)
        self.interface = []
        self.interface_index = {}
        self.notes_stack.append(self.notes)
        self.notes = {}
        for key in passed_notes:
            self.notes[key] = self.notes_stack[-1][key]
        
        logic_function, draw_function, setup_function = self.GetStateFunctions(state_name)
        self.logic_function = logic_function
        self.draw_function = draw_function
        self.setup_function = setup_function
        if self.setup_function:
            try:
                self.setup_function()
            except:
                raise "There was an error when running the setup function \""+self.setup_function.__name__+"\"."
        self.states.append(state_name)

    def PopState(self,passed_notes=[]):
        """End the current game state."""
        print "PopState. Remaining: "+str(self.states)

        if self.states:
            self.states.pop()
            if self.states:
                ## Revert to past interface elements & notes, passing some notes "down."
                self.interface = self.interface_stack.pop()
                self.interface_index = {}
                for widget in self.interface:
                    self.interface_index[widget.name] = widget
                for key in passed_notes:
                    self.notes_stack[-1][key] = self.notes[key]
                self.notes = self.notes_stack.pop()

                self.logic_function, self.draw_function, self.setup_function = self.GetStateFunctions(self.states[-1])
                if self.setup_function: ## Run it if it exists.
                    try:
                        self.setup_function()
                    except:
                        raise "There was an error when running the setup function \""+self.setup_function.__name__+"\"."

            else:
                self.logic_function, self.draw_function, setup_function = None, None, None

        else:
            raise "Error: There are no game states to pop!"

    def JumpState(self,state_name):
        """Pop and push to end this state and switch to another."""
        self.PopState()
        self.PushState(state_name)
        if self.setup_function:
            self.setup_function()

    def ClearState(self):
        """Pop all states."""
        self.states = self.states[:1]
        self.PopState()

    """ Main Loop """

    def MainLoop(self):
        """Do basic event-handling until the program ends.

        The game logic and drawing are handled by a pair
        of functions that can be replaced on the fly to
        represent switching to different aspects of gameplay,
        such as a title screen, battle screen etc.."""
        cycles = 0
        starting_time = time.time()
        while self.states: ## Until there's nothing to do:
            self.draw_function()
            unhandled_events = self.logic_function()
            self.messages = []

            if unhandled_events:
                for event in unhandled_events:
                    if event.type == QUIT:
                        return
                    elif event.type == KEYDOWN and event.key == K_ESC:
                        return

            pygame.display.update()
            self.clock.tick(self.framerate)
            cycles += 1
##            print cycles

        ## At program's end, display the overall framerate.
        if cycles > 1:
            ending_time = time.time()
            fps = cycles/float(ending_time - starting_time)
            print "FPS: "+str(fps)

    """ Other """

    def AddWidget(self,widget):
        """Adds to a list so that widgets can be drawn in a specific order."""
        self.interface.append(widget)
        self.interface_index[widget.name] = widget

    def ResetInterface(self):
        self.interface = []
        self.interface_index = {}

    def SetDesiredFramerate(self,fps=20):
        self.framerate = fps


class Demo( Framework ):
    """Demonstrate a very simple program using the above framework.

    All it does is run one state, stopping after a certain number of loops.
    In a game you would create a class like this and push/pop a larger
    set of states according to decisions made in the logic functions.
    Note that the trio of *Logic, *Draw and (optionally) *Setup conform to
    a naming standard, so that just invoking the state name "Test"
    makes them accessible.
    """
    def __init__(self,**options):
        Framework.__init__(self,**options)

    def TestSetup(self):
        self.notes["count"] = 0

    def TestLogic(self):
        count = self.notes["count"]
        print "Count: "+str(count)
        if count > 28:
            self.PopState()
        self.notes["count"] = count + 1

    def TestDraw(self):
        ## Draw a wierd little cross thing as a test.
        screen.fill((0,0,0))
        count = self.notes["count"]
        pygame.draw.rect(screen,(count*3,count*5,count*7),(150,150,100,100),count*10)


## Autorun (Demo)
if __name__ == "__main__":
    print "Running demo."
    import time
    import pygame
    screen = pygame.display.set_mode((400,400))
    d = Demo()
    d.PushState("Test")
    d.MainLoop()
