#!/usr/bin/python
"""
Trade concept test
by Kris Schnee.
"""

##___________________________________
##          ~Header~
##___________________________________

__author__ = "Kris Schnee"
__version__ = "2007.5.29"
__license__ = "Not offered for redistribution or modification at this time."


##___________________________________
##      ~Imported Modules~
##___________________________________

import random
random.seed("Porcupine Sea")

import pygame
from pygame.locals import *

import ringtale
import driftwood

##___________________________________
##            ~Constants~
##___________________________________

DEFAULT_INVENTORY = {"fish":99,"grape":50,"iron":1,"brick":5,"apple":20}
DEFAULT_TRADE_VALUES = {"pearl_necklace":35,
                        "iron":25,
                        "grape":5,
                        "brick":10,
                        "apple":3,
                        "emerald":30,
                        "bread":3,
                        "fish":3,
                        "meat":4,
                        "flour":2,
                        }

DEFAULT_TRIBE = "Crab Crown"

COMMENTS_ON_BARGAIN = [(-2.0,"How generous I am! What a bargain for you!"),
                       (-1.0,"I'll offer all this. Aren't I kind?"),
                       (-.25,"For you, a special offer."),
                       (-.10,"I'll give you a good deal."),
                       (.10,"I think this is fair; don't you?"),
                       (.25,"This is pretty good."),
                       (1.0,"I'm happy; are you?"),
                       (2.0,"Hmm, you'd be generous to accept this."),
                       (9999999.0,"Wow! Should I accept such a gift?"),
                    ]

##___________________________________
##            ~Classes~
##___________________________________

class Inventory( dict ):
    """A simple inventory system for adding/removing items.

    Treats items as fungible and carries no data about, eg., item type.
    Set max item types/quantity to zero if you want no limits."""
    def __init__(self,**options):
        contents = options.get("contents",{})
        dict.__init__(self,contents)
        self.max_item_types = options.get("max_item_types",15)
        self.max_item_quantity = options.get("max_item_quantity",99)

    def Clear(self):
        self.clear()

    def Add(self,item,howmany=1):
        """Returns True if successful, else returns False."""
        n = self.get(item,0)
        if self.max_item_quantity and n + howmany > self.max_item_quantity:
            return False
        elif self.max_item_types and item not in self and len(self) == self.max_item_types:
            return False
        else:
            self[item] = n + howmany
            return True

    def CanAdd(self,item,howmany):
        """Will an attempt to add this much succeed?"""
        n = self.get(item,0)
        if self.max_item_quantity and n + howmany > self.max_item_quantity:
            return False
        elif self.max_item_types and item not in self and len(self) == self.max_item_types:
            return False
        else:
            return True

    def AddThings(self,group):
        """Try to add a whole dictionary of things.

        Cancels and returns False if any addition is forbidden."""
        ## First, check whether this whole addition is okay. Don't add yet.
        for item in group:
            n = self.get(item,0)
            if self.max_item_quantity and n + group[item] > self.max_item_quantity:
                print "not OK: "+item+" (too much)"
                return False
            elif self.max_item_types and item not in self and len(self) == self.max_item_types:
                print "not OK: "+item+" (too many types)"
                return False

        ## If we're still here, the addition is okay.
        for item in group:
            self[item] = self.get(item,0) + group[item]
        return True

    def Remove(self,item,howmany=1):
        """Returns True if successful, else returns False."""
        n = self.get(item,0)
        if howmany > n:
            return False
        else:
            self[item] = n - howmany
            if not self[item]:
                del self[item]
            return True




class Trader:
    def __init__(self,**options):
        ## A bit about me.
        self.name = options.get("name","Trader")
        self.tribe = options.get("tribe",DEFAULT_TRIBE)
        self.age = 0 ## In days.

        ## My stock in trade.
        self.inventory = Inventory(contents=options.get("inventory",DEFAULT_INVENTORY))
        self.offer = Inventory(max_item_types=6) ## Stuff being "held out," not in inventory.

        ## My knowledge.
        self.values = options.get("values",DEFAULT_TRADE_VALUES.copy()) ## What I value.
        self.customers = {} ## Customers I've met.
        self.value_revision_rate = .03 ## Revise your values this fast.
        self.tradeback_reluctace = 50 ## % chance not to undo a purchase.

        ## Interaction with others.
        self.customer = None ## Who am I dealing with right now?
        self.comments = []
        self.opinion_of_deal = {}

        self.portrait = options.get("portrait")

        ## Unused
##        self.trade_percentage = .25
##        self.giver_dominance = .5

    def JudgeMyWealth(self):
        return self.Appraise(self.inventory)

    def WithdrawOffer(self):
        """Return all offered goods to my inventory."""
        self.inventory.AddThings(self.offer)
        self.offer.Clear()

    def Appraise(self,offer={},my_values=True):
        """Estimates the value of any set of goods (mine or yours)."""
        if my_values: ## I'll use my own value estimates.
            values = self.values
        else: ## I'll judge how much _you_ value these goods.
            if not self.customer:
                raise self.name+": You need to call StartTrading first."
            values = self.customer["values"]
        value = 0.0
        for item in offer:
            try:
                value += values[item] * offer[item]
            except:
                raise "Error: I don't have a value for "+item+"!"
        return value

    def StartTrading(self,customer_name):
        """Begin a trading session.

        If I know you, then I set self.customer so that the record is
        readily available. If not, then I make a new record.
        A key feature here is that I keep an estimate of what you
        consider valuable, based on your actions, and that's kept
        separately from my own set of values.
        """
        self.comments = []
        self.customer = self.customers.get(customer_name)
        if not self.customer:
            self.customer = {"name":customer_name,
                             "last_seen":self.age,
                             "values":self.values.copy(), ## Starting estimate.
                             "last_offer":{},
                             "already_traded":[],
                             "wants":[],
                             "does_not_want":[],
                             "wants_the_numbers":False,
                             }
            self.customers[customer_name] = self.customer
            self.comments.append("Good to meet you! What do you have?")
        else:
##            self.comment = "It's been "+str(self.world.time-self.customer["last_seen"])+" days, hasn't it?"
            self.comments.append("Welcome back to "+str(self.tribe)+"!")

    def StopTrading(self):
        self.customer = None

    def IndicateItemWanted(self,item,wanted=0):
        """Mark an item as especially wanted or unwanted."""
        if not self.customer:
            raise self.name+": You need to call StartTrading first."        

        if wanted == 1:
            if item in self.customer["does_not_want"]:
                self.customer["does_not_want"].remove(item)
                self.comments.append("Oh, you DO want "+item.title()+"?")
            if not item in self.customer["wants"]:
                self.customer["wants"].append(item)
        elif wanted == -1:
            if item in self.customer["wants"]:
                self.customer["wants"].remove(item)
                self.comments.append("You don't want "+item.title()+" after all?")
            if not item in self.customer["does_not_want"]:
                self.customer["does_not_want"].append(item)
            if item in self.offer:
                n = self.offer[item]
                self.offer.Remove(item,n)
                self.inventory.Add(item,n)
        else:
            if item in self.customer["wants"]:
                self.customer["wants"].remove(item)
            if item in self.customer["does_not_want"]:
                self.customer["does_not_want"].remove(item)

    def MakeOffer(self,your_offer):
        """Modify what I'm offering to match percieved value of your offer.

        That is, I'll change the array of goods in self.offer, so that
        my estimate of my offer's value _to you_ is comparable to my
        estimate of your offer's value _to me_.

        So far the commentary is handled haphazardly.
        """

        print "\n --- Making an offer. ---"

        self.comments = []

        self.events = {} ## Events to comment on. NYI.

        ## Get my estimate of what you value.
        if not self.customer:
            raise self.name+": You need to call StartTrading first."
        your_values = self.customer["values"]
        last_offer = self.customer["last_offer"]
        print "Your offer: "+str(your_offer)
        print "Last offer: "+str(last_offer)

        ## Note what was added/removed and change my estimate of your values.
        added = [(i,your_offer[i]-last_offer.get(i,0)) for i in your_offer if your_offer[i] > last_offer.get(i,0)]
        removed = [(i,last_offer[i]-your_offer.get(i,0)) for i in last_offer if last_offer[i] > your_offer.get(i,0)]
        for item in added:
            i, n = item
            print "You added: "+i+" x"+str(n)
            value_revision = pow(1.0-self.value_revision_rate,n)
            print "Revised my estimate of how much you value "+i+", from "+str(your_values[i])+" to "+str(your_values[i]*value_revision)[:5]+"."
            your_values[i] *= value_revision
        for item in removed:
            i, n = item
            print "You removed: "+i+" x"+str(n)
            value_revision = pow(1.0+self.value_revision_rate,n)
            print "Revised my estimate of how much you value "+i+", from "+str(your_values[i])+" to "+str(your_values[i]*value_revision)[:5]+"."
            your_values[i] *= value_revision

        ## I should also remove from my offer, any item you're offering me.
        for item in [entry[0] for entry in added if entry[0] in self.offer]:
            print "Removing from my offer: "+item
            n = self.offer[item]
            self.offer.Remove(item,n)
            self.inventory.Add(item,n)

        ## Did you hit "Offer" without putting up anything?
        if not your_offer:
            self.WithdrawOffer() ## "I'm not a charity, you know."
            if removed: ## You removed everything from your side.
                self.comments.append("Let's start over.")
            else:
                self.comments.append("Won't you offer anything?")
            return

        self.customer["last_offer"] = your_offer.copy() ## Note it for next time.

        ## What's your offer worth to me?
        your_offer_value = self.Appraise(your_offer,True)

        match_tolerance = .25 ## Try to come within this proportion of yours.
        attempts = 0
        max_attempts = 3

        satisfied = False
        while not satisfied:

            ## How much do I think my offer is worth to you?
            my_offer_value = self.Appraise(self.offer,False)
            print "Attempt #"+str(attempts)+". Your offer vs. mine: "+str((your_offer_value, my_offer_value))

            value_difference = abs(my_offer_value-your_offer_value)
            if value_difference <= your_offer_value * match_tolerance:
                print "Satisfied with this match."
                satisfied = True

            elif my_offer_value < your_offer_value:
                ## I should add something. What items can I add?
                possible_items = [item for item in self.inventory
                                  if self.inventory[item] > 0
                                  and item not in your_offer
                                  and item not in self.customer["does_not_want"]
                                  ]

                print "Checking something"
                print self.customer["does_not_want"]
                for i in possible_items:
                    print "Item: "+str(i)
                    assert i not in self.customer["does_not_want"]
                print "The items in possible_items are: "+str(possible_items)
                print "DNW: "+str(self.customer["does_not_want"])
                
                item = None

                ## Consider disqualifying items that you've given to me.
                for i in possible_items:
                    if i in self.customer["already_traded"]:
                        if random.randint(0,99) < self.tradeback_reluctace:
                            possible_items.remove(i)

                ## Considering above factors, is there even anything available?
                if not possible_items:
                    ## Can't add anything more! Bail.
                    print "Can't add more! Bailing."
                    self.comments.append("Sorry; I don't know what to offer!")
                    return self.offer

                ## Now, look at several methods of choosing an item.

                ## Is there a specific item you've requested?
                if self.customer["wants"]:
                    wants = [i for i in self.customer["wants"] if i in possible_items]
                    item = wants[random.randint(0,len(wants)-1)]
                    print "You've requested "+item+"."
                    self.comments.append("How about this "+item.title()+", then?")

                ## Is there something you'll like that I don't?
                good_items = [item for item in possible_items if your_values[item] > self.values[item]]
                if good_items:
                    item = good_items[random.randint(0,len(good_items)-1)]
                    print "I think you'll like "+item.title()+"."
                    self.comments.append("You like "+item.title()+", yes?")

                ## Still no item chosen? Pick one randomly then.
                if not item:
                    item = possible_items[random.randint(0,len(possible_items)-1)]
                    print "I'm guessing that you'll like this: "+item.title()

                print "Item selected: "+item

                ## How much of this item should I add?
                your_value_per_unit = your_values[item]
                ideal_quantity = max(1,int(value_difference/your_value_per_unit))
                quantity = min(self.inventory[item],
                               ideal_quantity)
                print "Ideal quantity: "+str(ideal_quantity)
                print "Quantity: "+str(quantity)
                add_okay = self.offer.Add(item,quantity)
                if add_okay:
                    self.inventory.Remove(item,quantity)
                    print "Added: "+item+" x"+str(quantity)
                    self.comments.append("I'll add "+str(quantity)+" "+item.title()+".")
                else:
                    print "Couldn't add "+item+" x"+str(quantity)

                assert item not in self.customer["does_not_want"]


            else:
                ## I'm offering too much. I should remove something.
##                satisfied = True ## NYI.
                ## Try a dead-simple method for now.
                item = self.offer.keys()[random.randint(0,len(self.offer)-1)]
                quantity = random.randint(1,self.offer[item])
                print "Withdrew: "+item+" x"+str(quantity)
                self.offer.Remove(item,quantity)
                self.inventory.Add(item,quantity)
                self.comments.append("Well then, I'll remove "+str(quantity)+" "+item.title()+".")

            attempts += 1
            if attempts == max_attempts:
                ## Bah, just give up and try this offer.
                satisfied = True

        if not self.comments: ## Nothing special to say? Describe what I think.
            self.JudgeDeal(your_offer,True)

        return self.offer

    def PrintOffer(self):
        print "Here's what I offer:"
        for item in self.offer:
            print item + " x "+str(self.offer[item])
        print "I appraise this stuff at "+str(self.Appraise(self.offer))+"."

    def JudgeDeal(self,your_offer,comment=True):
        """Returns an estimate of how good a deal is proposed.

        I judge both whether I think the deal is fair, considering
        how much _you_ seem to value what you're getting,
        and how much the deal benefits me.
        """
        value_given = self.Appraise(self.offer,True)
        value_gotten = self.Appraise(your_offer,True)
        value_you_give = self.Appraise(your_offer,False)
        value_you_get = self.Appraise(self.offer,False)

        deal_size = (value_given + value_gotten) / 2
        my_profit_margin = (value_gotten - value_given) / max(1,value_given)
        your_profit_margin = (value_you_get - value_you_give) / max(1,value_you_give)

        self.opinion_of_deal = {"deal_size":deal_size,
                                "my_profit_margin":my_profit_margin,
                                "your_profit_margin":your_profit_margin}

        if self.customer["wants_the_numbers"]:
            self.comments.append("Deal size: "+str(deal_size))
            self.comments.append("My profit margin: "+str(my_profit_margin*100)[:4]+"%")
            self.comments.append("Your profit margin: "+str(your_profit_margin*100)[:4]+"%")
        else:
            for pair in COMMENTS_ON_BARGAIN:
                if my_profit_margin < pair[0]:
                    self.comments.append(pair[1])
                    break

    def FinalizeDeal(self,goods_gotten):
        """Assumes you've already put my offered goods into your inventory."""
        okay = self.inventory.AddThings(goods_gotten)
        if okay:
            self.offer.Clear()
            for item in goods_gotten:
                if item not in self.customer["already_traded"]:
                    self.customer["already_traded"].append(item)

##        pc_status_change = bargain * self.giver_dominance
##        tribes.tribe["pc_status"] += pc_status_change

        ## Reset assumptions about goods customer wants.
        self.customer["does_not_want"] = []
        self.customer["wants"] = []

        return okay


##___________________________________
##            ~Functions~
##___________________________________


##__________________________________
##            ~Autorun~
##__________________________________

if __name__ == "__main__":
    t = Trader()



